use std::{collections::LinkedList, ops::Range};

use crate::token::Token;

use super::{Tree, ParseError, PartialTree, ParseErrorDetails};

#[derive(Default, Debug)]
pub struct Program(Vec<Declaration>, LinkedList<ParseError>);

#[derive(Default, Debug)]
pub struct PartialProgram(Program, Option<PartialDeclaration>);

impl PartialTree for PartialProgram {
	type Final = Program;

	fn accept(self, t: Token, s: Range<usize>) -> Result<(Self::Final, Token), Self> {
		match self.1 {
			Some(partial) =>
				// A partial parse is already in progress
				match partial.accept(t, s) {
					Ok((decl, tok)) => {
						// The parse finished, keep propegating the token
						self.0.0.push_front(decl);
						PartialProgram(self.0, None).accept(tok, s)
					}
					Err(partial) =>
						// The parse continues
						Err(PartialProgram(self.0, partial)),
				}
			None =>
				// We just finished parsing a declaration, and are now looking for a
				// DeclarationStart
				match t {
					Token::DeclarationStart =>
						// We got what we were looking for
						Err(PartialProgram(self.0, Some(Default::default()))),
					Token::Type |
					Token::Struct |
					Token::Trait |
					Token::Impl |
					Token::Symbol(_) => {
						// Seems like the user just forgot to start a declaration
						self.0.1.push_front(ParseError(
							s, ParseErrorDetails::MissingDeclarationStart
						));
						PartialProgram(self.0, Some(Default::default())).accept(t, s)
					}
					_ => {
						// No idea what's happening here, just throw it away
						self.0.1.push_front(ParseError(
							s, ParseErrorDetails::FoundGarbage(0)
						));
						Err(self)
					}
				}
		}
	}
}

impl Tree for Program {
	fn get_errors(&self) -> LinkedList<&ParseError> {
		self.0.iter()
			.map(|d| d.get_errors())
			.collect()
	}
}

impl Tree for PartialProgram {
	fn get_errors(&self) -> LinkedList<&ParseError> {
		self.0.get_errors()
	}
}
